//
// Written by Sander van Rossen
//	You are free to use this piece of code any way you like
//
#pragma once


/*! 
	@brief		Main template, this creates the function to convert an array 
				containing colors of type <tSrcColor> to colors of type 
				<tDstColor> and store those in another array..

	@warning	only use ColorDescription types with this structure!!
*/
template<typename tSrcColor, typename tDstColor>
struct ConvertColors
{
	enum EConversionConstants
	{
		//
		// The storage sizes of src & dst
		//
		SRC_STORAGE			= tSrcColor::COLOR_STORAGE_BITS,
		DST_STORAGE			= tDstColor::COLOR_STORAGE_BITS,


		//
		// The offset of each color component for src & dst
		//
		SRC_RED_OFFSET		= tSrcColor::RED_OFFSET		,
		SRC_GREEN_OFFSET	= tSrcColor::GREEN_OFFSET	,
		SRC_BLUE_OFFSET		= tSrcColor::BLUE_OFFSET	,
		SRC_ALPHA_OFFSET	= tSrcColor::ALPHA_OFFSET	,
		DST_RED_OFFSET		= tDstColor::RED_OFFSET		,
		DST_GREEN_OFFSET	= tDstColor::GREEN_OFFSET	,
		DST_BLUE_OFFSET		= tDstColor::BLUE_OFFSET	,
		DST_ALPHA_OFFSET	= tDstColor::ALPHA_OFFSET	,
		

		//
		// The sizes of each color component for src & dst
		//
		SRC_RED_SIZE		= tSrcColor::RED_SIZE		,
		SRC_GREEN_SIZE		= tSrcColor::GREEN_SIZE		,
		SRC_BLUE_SIZE		= tSrcColor::BLUE_SIZE		,
		SRC_ALPHA_SIZE		= tSrcColor::ALPHA_SIZE		,
		DST_RED_SIZE		= tDstColor::RED_SIZE		,
		DST_GREEN_SIZE		= tDstColor::GREEN_SIZE		,
		DST_BLUE_SIZE		= tDstColor::BLUE_SIZE		,
		DST_ALPHA_SIZE		= tDstColor::ALPHA_SIZE		,


		///
		/// The largest storage of the two (src/dst)
		///
		LARGEST_STORAGE		= (SRC_STORAGE > DST_STORAGE) 
								? DST_STORAGE : SRC_STORAGE,
		
		
		///
		/// Size of the most optimal number of color components to process at 
		/// the same time. 
		/// For example, converting 1bit colors to 2bit colors would require 
		/// us to process 4 colors at the same time (4 x 2bit == 8bit == 1byte)
		/// If we didn't do this, our calculations wouldn't necesarrily end on
		/// byte-boundaries, and make the calculations *a lot* more complex
		///
		CHUNK_SIZE			= (SRC_STORAGE > DST_STORAGE)				?	
								(SRC_STORAGE < BITS_PER_BYTE)			?
									(SRC_STORAGE / DST_STORAGE) 
										* (BITS_PER_BYTE / SRC_STORAGE)	:
									(SRC_STORAGE / DST_STORAGE)			:
								(DST_STORAGE < BITS_PER_BYTE)			?
									(DST_STORAGE / SRC_STORAGE)	
										* (BITS_PER_BYTE / DST_STORAGE)	:
									(DST_STORAGE / SRC_STORAGE)			,
		/// Lets make sure we process at least one color component per chunk
		CHUNK_LOOPS			= CHUNK_SIZE ? CHUNK_SIZE : 1,


		/// The size of each step to take in our outer-loop, 
		/// in bytes for the source data
		SRC_STEP			= (CHUNK_LOOPS * SRC_STORAGE) / BITS_PER_BYTE,
		
		/// The size of each step to take in our outer-loop, 
		/// in bytes for the destination data
		DST_STEP			= (CHUNK_LOOPS * DST_STORAGE) / BITS_PER_BYTE,
		
		/// The largest step of the two, to determine the required iterations
		/// of the loop
		LARGEST_STEP		= SRC_STEP > DST_STEP ? SRC_STEP : DST_STEP


	};//enum EConversionConstants



protected:
	/*! 
		@brief	This template unrolls the loop used to convert a color-'chunk'
				This 'chunk' is the most optimal number of colors to process 
				at once this is necesarry to keep calculations simple.
				If you don't do this you'll have a hard time converting 4 bit 
				pixels to 2 bit pixels simply because multiple source pixels 
				will fit into one 'destination' byte
	*/
	struct Chunk
	{
		template<cUInt tLoopNum, typename _unused_dummy_ = int>
		struct Loop
		{
			/*! 
				@brief	Some constants which are required for our calculations
			*/
			enum ELoopConstants
			{
				ELEMENT			= (tLoopNum - 1),
				DST_BIT_OFFSET	= (ELEMENT * DST_STORAGE) & (BITS_PER_INT-1),
				SRC_BYTE_OFFSET	= (ELEMENT * SRC_STORAGE) /  BITS_PER_BYTE,
				SRC_BIT_OFFSET	= (ELEMENT * SRC_STORAGE) & (BITS_PER_BYTE-1),
			};//enum ELoopConstants

			INLINE static cUInt CalcElement(rcpcUChar	inType,
											rcUInt		inSrcPos)
			{
				//
				// Convert a single color from our source to our destination
				//
				return
					  ConvertBits((SRC_RED_OFFSET	+       SRC_BIT_OFFSET ),
								   SRC_RED_SIZE		,
								  (DST_RED_OFFSET	+       DST_BIT_OFFSET ),
								   DST_RED_SIZE		,
								  (rUInt)(inType[inSrcPos + SRC_BYTE_OFFSET]))
					| ConvertBits((SRC_GREEN_OFFSET	+       SRC_BIT_OFFSET ),
								   SRC_GREEN_SIZE	,
								  (DST_GREEN_OFFSET	+       DST_BIT_OFFSET ),
			 					   DST_GREEN_SIZE	,
								  (rUInt)(inType[inSrcPos + SRC_BYTE_OFFSET]))
					| ConvertBits((SRC_BLUE_OFFSET	+       SRC_BIT_OFFSET ),
								   SRC_BLUE_SIZE	,
								  (DST_BLUE_OFFSET	+       DST_BIT_OFFSET ),
								   DST_BLUE_SIZE	,
								  (rUInt)(inType[inSrcPos + SRC_BYTE_OFFSET]))
					| ConvertBits((SRC_ALPHA_OFFSET	+       SRC_BIT_OFFSET ),
								   SRC_ALPHA_SIZE	,
								  (DST_ALPHA_OFFSET	+       DST_BIT_OFFSET ),
								   DST_ALPHA_SIZE	,
								  (rUInt)(inType[inSrcPos + SRC_BYTE_OFFSET])) 

				//
				// Add the next color to the destination...
				//
					| Loop<tLoopNum - 1,int>::CalcElement(inType, inSrcPos);
			}
		};//struct Loop


		/*! 
			@brief	Used to stop the template loop
		*/
		template<>
		struct Loop<0, int>
		{
			INLINE static cUInt CalcElement(rcpcUChar,rcUInt) { return 0; }
		};//struct Loop


	};//struct Chunk


public:
	/*! 
		@brief	Do the actual color conversion of the colors
	*/
	static void Convert(rcpUChar		outType,
						rcpcUChar		inType,
						rcUInt			inByteSize)
	{
		//
		// Divide the size of the source by the size of the 
		// conversion color chunks
		// (block of colors which are converted in one go)
		//
		cUInt	loop_size		= inByteSize / LARGEST_STEP;
		for (UInt counter = 0; counter < loop_size; counter++)
		{
			//
			// Calculate the source and destination positions of the input
			// and output arrays
			//
			cUInt		src_pos		= counter * SRC_STEP;
			cUInt		dst_pos		= counter * DST_STEP;

			//
			// Convert our chunk of colors
			//
			(rUInt)(outType[dst_pos]) = 
						Chunk::Loop<CHUNK_LOOPS>::
									CalcElement(inType, src_pos);
		}
	}


};//struct ConvertColors
